"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const bcrypt_1 = require("bcrypt");
const crypto_1 = require("crypto");
const ooth_i18n_1 = require("ooth-i18n");
const passport_local_1 = require("passport-local");
const util_1 = require("util");
const SALT_ROUNDS = 12;
const HOUR = 1000 * 60 * 60;
const DEFAULT_LANGUAGE = 'en';
const DEFAULT_TRANSLATIONS = {
    en: require('../i18n/en.json'),
};
const DEFAULT_VALIDATORS = {
    username: {
        regex: /^[a-z][0-9a-z_]{3,19}$/,
        error: 'validators.invalid_username',
    },
    password: {
        test: (password) => /\d/.test(password) && /[a-z]/.test(password) && /[A-Z]/.test(password) && /.{6,}/.test(password),
        error: 'validators.invalid_password',
    },
    email: {
        regex: /^.+@.+$/,
        error: 'validators.invalid_email',
    },
};
function randomToken() {
    return crypto_1.randomBytes(43).toString('hex');
}
function default_1({ name = 'local', ooth, defaultLanguage, translations, validators }) {
    const __ = ooth_i18n_1.getI18n(translations || DEFAULT_TRANSLATIONS, defaultLanguage || DEFAULT_LANGUAGE);
    const actualValidators = Object.assign({}, DEFAULT_VALIDATORS, validators);
    function testValue(key, value, language) {
        const test = actualValidators[key];
        if (test.regex) {
            if (!test.regex.test(value)) {
                throw new Error(__(test.error, null, language));
            }
        }
        else {
            if (!test.test(value)) {
                throw new Error(__(test.error, null, language));
            }
        }
    }
    ooth.registerUniqueField(name, 'username', 'username');
    ooth.registerUniqueField(name, 'email', 'email');
    ooth.registerProfileFields(name, 'username', 'email', 'verified');
    ooth.registerPrimaryAuth(name, 'login', [ooth.requireNotLogged], new passport_local_1.Strategy({
        usernameField: 'username',
        passwordField: 'password',
        passReqToCallback: true,
    }, util_1.callbackify((req, username, password) => __awaiter(this, void 0, void 0, function* () {
        const fullRequest = req;
        let user = yield ooth.getUserByUniqueField('username', username);
        if (!user) {
            user = yield ooth.getUserByUniqueField('email', username);
        }
        if (!user) {
            throw new Error(__('login.no_user', null, fullRequest.locale));
        }
        if (!user[name]) {
            throw new Error(__('login.no_password', null, fullRequest.locale));
        }
        if (!(yield bcrypt_1.compare(password, user[name].password))) {
            throw new Error(__('login.invalid_password', null, fullRequest.locale));
        }
        return user._id;
    }))));
    ooth.registerMethod(name, 'set-username', [ooth.requireLogged], ({ username }, userId, locale) => __awaiter(this, void 0, void 0, function* () {
        if (typeof username !== 'string') {
            throw new Error(__('set_username.invalid_username', null, locale));
        }
        testValue('username', username, locale);
        const existingUser = yield ooth.getUserByUniqueField('username', username);
        if (existingUser) {
            throw new Error(__('set_username.username_taken', null, locale));
        }
        yield ooth.updateUser(name, userId, {
            username,
        });
        return {
            message: __('set_username.username_updated', null, locale),
        };
    }));
    ooth.registerMethod(name, 'set-email', [ooth.requireLogged], ({ email }, userId, locale) => __awaiter(this, void 0, void 0, function* () {
        if (typeof email !== 'string') {
            throw new Error(__('set_email.invalid_email', null, locale));
        }
        testValue('email', email, locale);
        const existingUser = yield ooth.getUserByUniqueField('email', email);
        if (existingUser && existingUser._id !== userId) {
            throw new Error(__('set_email.email_already_registered', null, locale));
        }
        const verificationToken = randomToken();
        yield ooth.updateUser(name, userId, {
            email,
            verificationToken: yield bcrypt_1.hash(verificationToken, SALT_ROUNDS),
            verificationTokenExpiresAt: new Date(Date.now() + HOUR),
        });
        yield ooth.emit(name, 'set-email', {
            email,
            verificationToken,
            _id: userId,
        });
        return {
            message: __('set_email.email_updated', null, locale),
        };
    }));
    ooth.registerMethod(name, 'register', [ooth.requireNotLogged], ({ email, password }, _userId, locale) => __awaiter(this, void 0, void 0, function* () {
        if (typeof email !== 'string') {
            throw new Error(__('register.invalid_email', null, locale));
        }
        if (typeof password !== 'string') {
            throw new Error(__('register.invalid_password', null, locale));
        }
        testValue('password', password, locale);
        let verificationToken;
        const existingUser = yield ooth.getUserByUniqueField('email', email);
        if (existingUser) {
            throw new Error(__('register.email_already_registered', null, locale));
        }
        verificationToken = randomToken();
        const _id = yield ooth.insertUser(name, {
            email,
            password: yield bcrypt_1.hash(password, SALT_ROUNDS),
            verificationToken: yield bcrypt_1.hash(verificationToken, SALT_ROUNDS),
            verificationTokenExpiresAt: new Date(Date.now() + HOUR),
        });
        yield ooth.emit(name, 'register', {
            _id,
            email,
            verificationToken,
        });
        return {
            message: __('register.registered', null, locale),
        };
    }));
    ooth.registerMethod(name, 'generate-verification-token', [ooth.requireRegisteredWith(name)], (_, userId, locale) => __awaiter(this, void 0, void 0, function* () {
        const verificationToken = randomToken();
        const user = yield ooth.getUserById(userId);
        if (!user[name] || !user[name].email) {
            throw new Error(__('generate_verification_token.no_email', null, locale));
        }
        yield ooth.updateUser(name, userId, {
            verificationToken: yield bcrypt_1.hash(verificationToken, SALT_ROUNDS),
            verificationTokenExpiresAt: new Date(Date.now() + HOUR),
        });
        yield ooth.emit(name, 'generate-verification-token', {
            verificationToken,
            _id: userId,
            email: user[name].email,
        });
        return {
            message: __('generate_verification_token.token_generated', null, locale),
        };
    }));
    ooth.registerMethod(name, 'verify', [], ({ userId, token }, _userId, locale) => __awaiter(this, void 0, void 0, function* () {
        if (!userId) {
            throw new Error(__('verify.no_user_id', null, locale));
        }
        if (!token) {
            throw new Error(__('verify.token_generated', null, locale));
        }
        const user = yield ooth.getUserById(userId);
        if (!user) {
            throw new Error(__('verify.no_user', null, locale));
        }
        const strategyValues = user[name];
        if (!strategyValues || !strategyValues.email) {
            // No email to verify, but let's not leak this information
            throw new Error(__('verify.no_email', null, locale));
        }
        if (!(yield bcrypt_1.compare(token, strategyValues.verificationToken))) {
            throw new Error(__('verify.invalid_token', null, locale));
        }
        if (!strategyValues.verificationTokenExpiresAt) {
            throw new Error(__('verify.no_expiry', null, locale));
        }
        if (new Date() >= strategyValues.verificationTokenExpiresAt) {
            throw new Error(__('verify.expired_token', null, locale));
        }
        yield ooth.updateUser(name, user._id, {
            verified: true,
            verificationToken: null,
        });
        const newUser = yield ooth.getUserById(user._id);
        yield ooth.emit(name, 'verify', {
            _id: newUser._id,
            email: newUser[name].email,
        });
        return {
            message: __('verify.verified', null, locale),
        };
    }));
    ooth.registerMethod(name, 'forgot-password', [ooth.requireNotLogged], ({ username }, _userId, locale) => __awaiter(this, void 0, void 0, function* () {
        if (!username || typeof username !== 'string') {
            throw new Error(__('forgot_password.invalid_username', null, locale));
        }
        let user = yield ooth.getUserByUniqueField('username', username);
        if (!user) {
            user = yield ooth.getUserByUniqueField('email', username);
        }
        if (!user) {
            throw new Error(__('forgot_password.no_user', null, locale));
        }
        const email = ooth.getUniqueField(user, 'email');
        const passwordResetToken = randomToken();
        yield ooth.updateUser(name, user._id, {
            email,
            passwordResetToken: yield bcrypt_1.hash(passwordResetToken, SALT_ROUNDS),
            passwordResetTokenExpiresAt: new Date(Date.now() + HOUR),
        });
        yield ooth.emit(name, 'forgot-password', {
            email,
            passwordResetToken,
            _id: user._id,
        });
        return {
            message: __('forgot_password.token_generated', null, locale),
        };
    }));
    ooth.registerMethod(name, 'reset-password', [ooth.requireNotLogged], ({ userId, token, newPassword }, _userId, locale) => __awaiter(this, void 0, void 0, function* () {
        if (!userId) {
            throw new Error(__('reset_password.no_user_id', null, locale));
        }
        if (!token) {
            throw new Error(__('reset_password.no_token', null, locale));
        }
        if (!newPassword || typeof newPassword !== 'string') {
            throw new Error(__('reset_password.invalid_password', null, locale));
        }
        testValue('password', newPassword, locale);
        const user = yield ooth.getUserById(userId);
        if (!user) {
            throw new Error('User does not exist.');
        }
        const strategyValues = user[name];
        if (!strategyValues || !strategyValues.passwordResetToken) {
            throw new Error(__('reset_password.no_reset_token', null, locale));
        }
        if (!(yield bcrypt_1.compare(token, strategyValues.passwordResetToken))) {
            throw new Error(__('reset_password.invalid_token', null, locale));
        }
        if (!strategyValues.passwordResetTokenExpiresAt) {
            throw new Error(__('reset_password.no_expiry', null, locale));
        }
        if (new Date() >= strategyValues.passwordResetTokenExpiresAt) {
            throw new Error(__('reset_password.expired_token', null, locale));
        }
        yield ooth.updateUser(name, user._id, {
            passwordResetToken: null,
            password: yield bcrypt_1.hash(newPassword, SALT_ROUNDS),
        });
        yield ooth.emit(name, 'reset-password', {
            _id: user._id,
            email: strategyValues.email,
        });
        return {
            message: __('reset_password.password_reset', null, locale),
        };
    }));
    ooth.registerMethod(name, 'change-password', [ooth.requireLogged], ({ password, newPassword }, userId, locale) => __awaiter(this, void 0, void 0, function* () {
        if (typeof password !== 'string') {
            throw new Error(__('change_password.invalid_password', null, locale));
        }
        testValue('password', newPassword, locale);
        const user = yield ooth.getUserById(userId);
        const strategyValues = user[name];
        if ((password || (strategyValues && strategyValues.password)) && !(yield bcrypt_1.compare(password, strategyValues.password))) {
            throw new Error(__('change_password.invalid_password', null, locale));
        }
        yield ooth.updateUser(name, userId, {
            passwordResetToken: null,
            password: yield bcrypt_1.hash(newPassword, SALT_ROUNDS),
        });
        yield ooth.emit(name, 'change-password', {
            _id: userId,
            email: strategyValues && strategyValues.email,
        });
        return {
            message: __('change_password.password_changed', null, locale),
        };
    }));
}
exports.default = default_1;
